@@ -9,7 +9,7 @@ import traceback
 import _thread
 from datetime import datetime
 from functools import partial
-from threading import Thread
+from threading import current_thread, main_thread, Thread
 from typing import Any, Callable, List, Optional, Tuple, Union, Type
 
 try:
@@ -84,7 +84,7 @@ class AbstractWorker:
         self.is_running_lock = ctx.Lock()
 
         # Register handler for graceful shutdown. This doesn't work on Windows
-        if not RUNNING_WINDOWS:
+        if not RUNNING_WINDOWS and current_thread() == main_thread():
             signal.signal(signal.SIGUSR1, self._exit_gracefully)
 
     def _exit_gracefully(self, *_) -> None:
@@ -107,7 +107,7 @@ class AbstractWorker:
                 self.is_running = False
                 raise StopWorker
 
-    def _exit_gracefully_windows(self):
+    def _exit_gracefully_windows(self) -> None:
         """
         Windows doesn't fully support signals as Unix-based systems do. Therefore, we have to work around it. This
         function is started in a thread. We wait for a kill signal (Event object) and interrupt the main thread if we
@@ -129,12 +129,13 @@ class AbstractWorker:
         life span is not yet reached it will execute the new task and put the results in the results queue.
         """
         # Enable graceful shutdown for Windows. Note that we can't kill threads in Python
-        if RUNNING_WINDOWS and self.pool_params.start_method != "threading":
+        if RUNNING_WINDOWS and self.pool_params.start_method != "threading" and current_thread() == main_thread():
             signal.signal(signal.SIGINT, self._exit_gracefully)
             t = Thread(target=self._exit_gracefully_windows)
             t.start()
 
-        self.worker_comms.signal_worker_alive(self.worker_id)
+        with self.worker_comms.get_worker_dead_lock(self.worker_id):
+            self.worker_comms.signal_worker_alive(self.worker_id)
 
         # Set tqdm and dashboard connection details. This is needed for nested pools and in the case forkserver or
         # spawn is used as start method
@@ -181,7 +182,7 @@ class AbstractWorker:
                         if self.map_params.worker_exit:
                             self._run_exit_func(additional_args)
                             self.worker_comms.wait_until_all_exit_results_obtained()
-                        if self.worker_comms.has_progress_bar():
+                        if self.map_params.progress_bar:
                             self.worker_comms.wait_until_progress_bar_is_complete()
                         return
                     else:
@@ -235,7 +236,8 @@ class AbstractWorker:
                 self.worker_comms.signal_worker_restart(self.worker_id)
 
         finally:
-            self.worker_comms.signal_worker_dead(self.worker_id)
+            with self.worker_comms.get_worker_dead_lock(self.worker_id):
+                self.worker_comms.signal_worker_dead(self.worker_id)
 
     def _get_func(self, additional_args: List) -> Callable:
         """
@@ -259,7 +261,17 @@ class AbstractWorker:
             with TimeIt(self.worker_insights.worker_init_time, self.worker_id):
                 self.map_params.worker_init(*additional_args)
 
-        return self._run_safely(_init_func, no_args=True)[1]
+        # Optionally update timeout info
+        if self.map_params.worker_init_timeout is not None:
+            try:
+                self.worker_comms.signal_worker_init_started(self.worker_id)
+                should_return = self._run_safely(_init_func, no_args=True)[1]
+            finally:
+                self.worker_comms.signal_worker_init_completed(self.worker_id)
+        else:
+            should_return = self._run_safely(_init_func, no_args=True)[1]
+
+        return should_return
 
     def _run_func(self, func: Callable, args: List) -> Tuple[Any, bool]:
         """
@@ -273,11 +285,21 @@ class AbstractWorker:
         def _func():
             with TimeIt(self.worker_insights.worker_working_time, self.worker_id,
                         self.max_task_duration_list, lambda: self._format_args(args, separator=' | ')):
-                results = func(args)
+                _results = func(args)
             self.worker_insights.update_n_completed_tasks(self.worker_id)
-            return results
+            return _results
 
-        return self._run_safely(_func, args)
+        # Optionally update timeout info
+        if self.map_params.task_timeout is not None:
+            try:
+                self.worker_comms.signal_worker_task_started(self.worker_id)
+                results = self._run_safely(_func, args)
+            finally:
+                self.worker_comms.signal_worker_task_completed(self.worker_id)
+        else:
+            results = self._run_safely(_func, args)
+
+        return results
 
     def _run_exit_func(self, additional_args: List) -> bool:
         """
@@ -290,7 +312,15 @@ class AbstractWorker:
             with TimeIt(self.worker_insights.worker_exit_time, self.worker_id):
                 return self.map_params.worker_exit(*additional_args)
 
-        results, should_return = self._run_safely(_exit_func, no_args=True)
+        # Optionally update timeout info
+        if self.map_params.worker_exit_timeout is not None:
+            try:
+                self.worker_comms.signal_worker_exit_started(self.worker_id)
+                results, should_return = self._run_safely(_exit_func, no_args=True)
+            finally:
+                self.worker_comms.signal_worker_exit_completed(self.worker_id)
+        else:
+            results, should_return = self._run_safely(_exit_func, no_args=True)
 
         if should_return:
             return True
@@ -327,7 +357,7 @@ class AbstractWorker:
                 # The main process tells us to stop working, shutting down
                 raise
 
-            except Exception as err:
+            except (Exception, SystemExit) as err:
                 # An exception occurred inside the provided function. Let the signal handler know it shouldn't raise any
                 # StopWorker exceptions from the parent process anymore, we got this.
                 with self.is_running_lock:
@@ -344,7 +374,7 @@ class AbstractWorker:
         # Carry on
         return results, False
 
-    def _raise(self, args: Any, no_args: bool, err: Exception) -> None:
+    def _raise(self, args: Any, no_args: bool, err: Union[Exception, SystemExit]) -> None:
         """
         Create exception and pass it to the parent process. Let other processes know an exception is set
 
@@ -380,7 +410,7 @@ class AbstractWorker:
 
                 # Add exception. When we have a progress bar, we add an additional one
                 self.worker_comms.add_exception(type(err), traceback_str)
-                if self.worker_comms.has_progress_bar():
+                if self.map_params.progress_bar:
                     self.worker_comms.add_exception(type(err), traceback_str)
 
     def _format_args(self, args: Any, no_args: bool = False, separator: str = '\n') -> str:
@@ -447,7 +477,7 @@ class AbstractWorker:
 
         :param force_update: Whether to force an update
         """
-        if self.worker_comms.has_progress_bar():
+        if self.map_params.progress_bar:
             (self.progress_bar_last_updated,
              self.progress_bar_n_tasks_completed) = self.worker_comms.task_completed_progress_bar(
                 self.worker_id, self.progress_bar_last_updated, self.progress_bar_n_tasks_completed, force_update
