@@ -1,10 +1,12 @@
 import logging
 import os
+import time
 import types
 import unittest
 import warnings
 from itertools import product, repeat
 from multiprocessing import Barrier, Value
+from threading import current_thread, main_thread, Thread
 from unittest.mock import patch
 
 import numpy as np
@@ -25,7 +27,7 @@ def square(idx, x):
     return idx, x * x
 
 
-def extremely_large_output(idx, x):
+def extremely_large_output(idx, _):
     return idx, os.urandom(1024 * 1024)
 
 
@@ -234,6 +236,38 @@ class MapTest(unittest.TestCase):
                 self.assertEqual(self.test_desired_output, sorted(results_list, key=lambda tup: tup[0]))
 
 
+class PoolInThreadTest(unittest.TestCase):
+
+    def setUp(self):
+        self.test_data = [1, 2, 3, 5, 6, 9, 37, 42, 1337, 0, 3, 5, 0]
+        self.test_desired_output = [self._square(x) for x in self.test_data]
+
+    def test_start_methods(self):
+        """
+        Test that a WorkerPool can be started inside a thread, which isn't the main thread. Test for different start
+        methods. All should work just fine
+        """
+        for start_method in TEST_START_METHODS:
+            with self.subTest(start_method=start_method):
+                t = Thread(target=self._map_thread, args=(start_method,))
+                t.start()
+                t.join()
+
+    def _map_thread(self, start_method):
+        """
+        This function is called from within a thread
+        """
+        self.assertNotEqual(current_thread(), main_thread())
+        with WorkerPool(2, start_method=start_method) as pool:
+            results_list = pool.map(self._square, self.test_data)
+            self.assertIsInstance(results_list, list)
+            self.assertListEqual(self.test_desired_output, results_list)
+
+    @staticmethod
+    def _square(x):
+        return x * x
+
+
 class WorkerIDTest(unittest.TestCase):
 
     def test_by_config_function(self):
@@ -680,20 +714,14 @@ class DaemonTest(unittest.TestCase):
         Test for different start methods
         """
         for start_method in TEST_START_METHODS:
-            if start_method == 'threading':
-                continue
             with self.subTest(start_method=start_method), \
                     WorkerPool(n_jobs=2, daemon=False, start_method=start_method) as pool:
                 pool.map(self._square_daemon, ((X,) for X in repeat(self.test_data, 3)), chunk_size=1)
 
-        # This won't work for threading
-        with self.assertRaises(ValueError), WorkerPool(n_jobs=3, daemon=False, start_method='threading') as pool:
-            pool.map(self._square_daemon, ((X,) for X in repeat(self.test_data, 3)), chunk_size=1)
-
     @staticmethod
-    def _square_daemon(X):
+    def _square_daemon(x):
         with WorkerPool(n_jobs=4) as pool:
-            return pool.map(square, X, chunk_size=1)
+            return pool.map(square, x, chunk_size=1)
 
 
 class CPUPinningTest(unittest.TestCase):
@@ -722,7 +750,7 @@ class CPUPinningTest(unittest.TestCase):
                                                (4, [[0, 3]], [[0, 3], [0, 3], [0, 3], [0, 3]])]:
             # The test has been designed for a system with at least 4 cores. We'll skip those test cases where the CPU
             # IDs exceed the number of CPUs.
-            if cpu_ids is not None and np.array(cpu_ids).max() >= cpu_count():
+            if cpu_ids is not None and np.array(cpu_ids).max(initial=0) >= cpu_count():
                 continue
 
             with self.subTest(n_jobs=n_jobs, cpu_ids=cpu_ids), patch('mpire.pool.set_cpu_affinity') as p, \
@@ -1156,6 +1184,52 @@ class ExceptionTest(unittest.TestCase):
                 with self.subTest(function='square_raises_on_idx', map='imap'), self.assertRaises(ValueError):
                     list(pool.imap_unordered(self._square_raises_on_idx, self.test_data, progress_bar=progress_bar))
 
+    def test_defunct_processes_exit(self):
+        """
+        Tests if MPIRE correctly shuts down after process becomes defunct using exit()
+        """
+        print()
+        for n_jobs, progress_bar, worker_lifespan in [(1, False, None),
+                                                      (3, True, 1),
+                                                      (3, False, 3)]:
+            for start_method in TEST_START_METHODS:
+                # Progress bar on Windows + threading is not supported right now
+                if RUNNING_WINDOWS and start_method == 'threading' and progress_bar:
+                    continue
+                self.logger.debug(f"========== {start_method}, {n_jobs}, {progress_bar}, {worker_lifespan} ==========")
+                with self.subTest(n_jobs=n_jobs, progress_bar=progress_bar, worker_lifespan=worker_lifespan,
+                                  start_method=start_method), self.assertRaises(SystemExit), \
+                        WorkerPool(n_jobs=n_jobs, start_method=start_method) as pool:
+                    pool.map(self._exit, range(100), progress_bar=progress_bar, worker_lifespan=worker_lifespan)
+
+    def test_defunct_processes_kill(self):
+        """
+        Tests if MPIRE correctly shuts down after one process becomes defunct using os.kill().
+
+        We kill worker 0 and to be sure it's alive we set an event object and then go in an infinite loop. The kill
+        thread waits until the event is set and then kills the worker. The other workers are also ensured to have done
+        something so we can test what happens during restarts
+        """
+        print()
+        for n_jobs, progress_bar, worker_lifespan in [(1, False, None),
+                                                      (3, True, 1),
+                                                      (3, False, 3)]:
+            for start_method in TEST_START_METHODS:
+                # Can't kill threads
+                if start_method == 'threading':
+                    continue
+
+                self.logger.debug(f"========== {start_method}, {n_jobs}, {progress_bar}, {worker_lifespan} ==========")
+                with self.subTest(n_jobs=n_jobs, progress_bar=progress_bar, worker_lifespan=worker_lifespan,
+                                  start_method=start_method), self.assertRaises(RuntimeError), \
+                        WorkerPool(n_jobs=n_jobs, pass_worker_id=True, start_method=start_method) as pool:
+                    events = [pool.ctx.Event() for _ in range(n_jobs)]
+                    kill_thread = Thread(target=self._kill_process, args=(events[0], pool))
+                    kill_thread.start()
+                    pool.set_shared_objects(events)
+                    pool.map(self._worker_0_sleeps_others_square, range(1000), progress_bar=progress_bar,
+                             worker_lifespan=worker_lifespan, chunk_size=1)
+
     @staticmethod
     def _square_raises(_, x):
         raise ValueError(x)
@@ -1166,3 +1240,120 @@ class ExceptionTest(unittest.TestCase):
             raise ValueError(x)
         else:
             return idx, x * x
+
+    @staticmethod
+    def _exit(_):
+        exit()
+
+    @staticmethod
+    def _worker_0_sleeps_others_square(worker_id, events, x):
+        """
+        Worker 0 waits until the other workers have at least spun up and then sets her event and sleeps
+        """
+        if worker_id == 0:
+            [event.wait() for event in events[1:]]
+            events[0].set()
+            while True:
+                pass
+        else:
+            events[worker_id].set()
+            return x * x
+
+    @staticmethod
+    def _kill_process(event, pool):
+        """
+        Wait for event and kill
+        """
+        event.wait()
+        pool._workers[0].terminate()
+
+
+class TimeoutTest(unittest.TestCase):
+
+    def setUp(self):
+        # Create some test data
+        self.test_data = [1, 2, 3]
+
+    def test_worker_init_timeout(self):
+        """
+        Checks if the worker_init timeout is properly triggered
+        """
+        for start_method in TEST_START_METHODS:
+
+            with self.subTest('Well below timeout', start_method=start_method), \
+                    WorkerPool(2, start_method=start_method) as pool:
+                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1,
+                                              worker_init_timeout=100), self.test_data)
+
+            with self.subTest('Exceeding timeout, map', start_method=start_method), \
+                    WorkerPool(2, start_method=start_method) as pool, self.assertRaises(TimeoutError):
+                pool.map(self._f1, self.test_data, worker_init=self._init2, worker_init_timeout=0.1)
+
+            with self.subTest('Exceeding timeout, imap', start_method=start_method), \
+                    WorkerPool(2, start_method=start_method) as pool, self.assertRaises(TimeoutError):
+                for _ in pool.imap(self._f1, self.test_data, worker_init=self._init2, worker_init_timeout=0.1):
+                    pass
+
+    def test_worker_task_timeout(self):
+        """
+        Checks if the worker_init timeout is properly triggered
+        """
+        for start_method in TEST_START_METHODS:
+
+            with self.subTest('Well below timeout', start_method=start_method), \
+                    WorkerPool(2, start_method=start_method) as pool:
+                self.assertListEqual(pool.map(self._f1, self.test_data, task_timeout=100), self.test_data)
+
+            with self.subTest('Exceeding timeout, map', start_method=start_method), \
+                    WorkerPool(2, start_method=start_method) as pool, self.assertRaises(TimeoutError):
+                pool.map(self._f2, self.test_data, task_timeout=0.1)
+
+            with self.subTest('Exceeding timeout, imap', start_method=start_method), \
+                    WorkerPool(2, start_method=start_method) as pool, self.assertRaises(TimeoutError):
+                for _ in pool.imap(self._f2, self.test_data, task_timeout=0.1):
+                    pass
+
+    def test_worker_exit_timeout(self):
+        """
+        Checks if the worker_exit timeout is properly triggered
+        """
+        for start_method in TEST_START_METHODS:
+
+            with self.subTest('Well below timeout', start_method=start_method), \
+                    WorkerPool(2, start_method=start_method) as pool:
+                self.assertListEqual(pool.map(self._f1, self.test_data, worker_exit=self._exit1,
+                                              worker_exit_timeout=100), self.test_data)
+
+            with self.subTest('Exceeding timeout, map', start_method=start_method), \
+                    WorkerPool(2, start_method=start_method) as pool, self.assertRaises(TimeoutError):
+                pool.map(self._f1, self.test_data, worker_exit=self._exit2, worker_exit_timeout=0.1)
+
+            with self.subTest('Exceeding timeout, imap', start_method=start_method), \
+                    WorkerPool(2, start_method=start_method) as pool, self.assertRaises(TimeoutError):
+                for _ in pool.imap(self._f1, self.test_data, worker_exit=self._exit2, worker_exit_timeout=0.1):
+                    pass
+
+    @staticmethod
+    def _init1():
+        pass
+
+    @staticmethod
+    def _init2():
+        time.sleep(1)
+
+    @staticmethod
+    def _f1(x):
+        return x
+
+    @staticmethod
+    def _f2(x):
+        time.sleep(1)
+        return x
+
+    @staticmethod
+    def _exit1():
+        pass
+
+    @staticmethod
+    def _exit2():
+        time.sleep(1)
